from dataclasses import dataclass, field
from enum import Enum
from typing import List
from dataclasses_json import dataclass_json, Undefined, CatchAll

from weatherflow4py.const import BASE_LOGGER
from weatherflow4py.models.rest.forecast import WindDirection


class WetBulbFlag(Enum):
    NONE = 0
    WHITE = 1
    GREEN = 2
    YELLOW = 3
    RED = 4
    BLACK = 5


@dataclass_json
@dataclass(frozen=True, eq=True)
class Observation:
                
    air_temperature: float

    brightness: int

    dew_point: float
    feels_like: float
    heat_index: float

    precip: float
    pressure_trend: str
    relative_humidity: int

    solar_radiation: int

    timestamp: int
    uv: float


    wind_avg: float
    wind_chill: float
    wind_direction: int
    wind_gust: float
    wind_lull: float

    # Optional Fields Discovered June 18
    air_density: float | None = None
    barometric_pressure: float | None = None
    delta_t: float | None = None
    sea_level_pressure: float | None = None
    station_pressure: float | None = None
    wet_bulb_globe_temperature: float | None = None
    wet_bulb_temperature: float | None = None


    # Potentially optional fields based on test data - this may grow over time.
    precip_accum_last_1hr: float = field(default=0.0)
    precip_accum_local_day: float = field(default=0.0)
    precip_accum_local_day_final: float = field(default=0.0)
    precip_accum_local_yesterday: float = field(default=0.0)
    precip_accum_local_yesterday_final: float = field(default=0.0)
    precip_analysis_type_yesterday: int = field(default=0)
    precip_minutes_local_day: int = field(default=0)
    precip_minutes_local_yesterday: int = field(default=0)
    precip_minutes_local_yesterday_final: int = field(default=0)

    lightning_strike_count: int = field(default=0)
    lightning_strike_count_last_1hr: int = field(default=0)
    lightning_strike_count_last_3hr: int = field(default=0)
    lightning_strike_last_distance: int = field(default=0)
    lightning_strike_last_epoch: int | None = field(default=None)

    @property
    def precip_accum_local_day_nearcast(self) -> float:
        """Alias for nearcast"""
        return self.precip_accum_local_day_final
    @property
    def precip_accum_local_yesterday_nearcast(self) -> float:
        """Alias for nearcast"""
        return self.precip_accum_local_yesterday_final

    @property
    def precip_minutes_local_yesterday_nearcast(self) -> float:
        """Alias for nearcast"""
        return self.precip_minutes_local_yesterday_final

    @property
    def wind_cardinal_direction(self) -> str:
        dirs = [
            WindDirection.N,
            WindDirection.NNE,
            WindDirection.NE,
            WindDirection.ENE,
            WindDirection.E,
            WindDirection.ESE,
            WindDirection.SE,
            WindDirection.SSE,
            WindDirection.S,
            WindDirection.SSW,
            WindDirection.SW,
            WindDirection.WSW,
            WindDirection.W,
            WindDirection.WNW,
            WindDirection.NW,
            WindDirection.NNW,
        ]
        ix = round(self.wind_direction / (360.0 / len(dirs)))
        return dirs[ix % len(dirs)].value

    @property
    def wet_bulb_globe_temperature_category(self) -> int:
        """Calculates a wet bulb globe temp category - based on wikipedia:
        https://en.wikipedia.org/wiki/Wet-bulb_globe_temperature."""
        if self.wet_bulb_globe_temperature <= 25.6:
            return 0  # No Risk
        if self.wet_bulb_globe_temperature <= 27.7:
            return 1  # Low risk
        if self.wet_bulb_globe_temperature <= 29.4:
            return 2  # Moderate risk
        if self.wet_bulb_globe_temperature <= 31.0:
            return 3
        if self.wet_bulb_globe_temperature <= 32.1:
            return 4
        return 5

    @property
    def wet_bulb_globe_temperature_flag(self) -> WetBulbFlag:
        return WetBulbFlag(self.wet_bulb_globe_temperature_category)

    @property
    def uv_index_color(self) -> str:
        """Return the UV index color."""
        if self.uv <= 2:
            return "green"
        if self.uv <= 5:
            return "yellow"
        if self.uv <= 7:
            return "orange"
        if self.uv <= 10:
            return "red"
        return "purple"

    @property
    def uv_index_exposure(self) -> str:
        """Return the UV exposure level."""
        if self.uv <= 2:
            return "low"
        if self.uv <= 5:
            return "moderate"
        if self.uv <= 7:
            return "high"
        if self.uv <= 10:
            return "very high"
        return "extreme"


@dataclass_json
@dataclass(frozen=True, eq=True)
class StationUnits:
    """The station_units values represent the units of the Station's owner, not the units of the observation values
    in the API response."""

    units_temp: str
    units_wind: str
    units_precip: str
    units_pressure: str
    units_distance: str
    units_direction: str
    units_other: str


@dataclass_json
@dataclass(frozen=True, eq=True)
class StationStatus:
    status_code: int
    status_message: str

    def __post_init__(self):
        if (
            self.status_message
            == "SUCCESS - Either no capabilities or no recent observations"
        ):
            BASE_LOGGER.debug(
                "No Capabilities or Recent Observations - Station may be offline"
            )


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass(frozen=True, eq=True)
class ObservationStationREST:
    """Top level field for both the observations/device/device_id and observations/station/station_id endpoints."""

    elevation: float
    is_public: bool
    latitude: float
    longitude: float
    obs: List[Observation]
    outdoor_keys: List[str]
    public_name: str
    station_id: int
    station_name: str
    station_units: StationUnits
    status: StationStatus
    timezone: str
    unknown_fields: CatchAll



"""Custom websocket types."""
from enum import Enum


class ObservationType(Enum):
    """Enumeration of the different observation types."""

    OBS_SKY = "obs_sky"
    OBS_ST = "obs_st"
    OBS_AIR = "obs_air"


class PrecipitationType(Enum):
    """(0 = none, 1 = rain, 2 = hail, 3 = rain + hail (experimental))"""

    NONE = 0
    RAIN = 1
    HAIL = 2
    RAIN_AND_HAIL = 3


class PrecipitationAnalysisType(Enum):
    """0 = none, 1 = Rain Check with user display on, 2 = Rain Check with user display off"""

    NONE = 0
    RAIN_CHECK_WITH_USER_DISPLAY_ON = 1
    RAIN_CHECK_WITH_USER_DISPLAY_OFF = 2



from enum import Enum


class EventType(Enum):
    ACKNOWLEDGEMENT = "ack"
    RAPID_WIND = "rapid_wind"
    OBSERVATION = "obs_st"
    LIGHTNING_STRIKE = "evt_strike"
    RAIN = "evt_precip"
    DEVICE_STATUS = "device_status"
    INVALID = "unknown"


class MessageType(Enum):
    LISTEN_START = "listen_start"
    LISTEN_STOP = "listen_stop"
    RAPID_WIND_START = "listen_rapid_start"
    RAPID_WIND_STOP = "listen_rapid_stop"


class LightingStrikeType:
    ALL = "all"
    CLOUD_TO_GROUND = "cg"
    CLOUD_TO_CLOUD = "ic"



"""Model for the forecast endpoint."""

from dataclasses import dataclass
from datetime import datetime
from typing import List
from dataclasses_json import dataclass_json
from enum import Enum


class Condition(Enum):
    CLEAR = "Clear"
    RAIN_LIKELY = "Rain Likely"
    RAIN_POSSIBLE = "Rain Possible"
    SNOW = "Snow"
    SNOW_LIKELY = "Snow Likely"
    SNOW_POSSIBLE = "Snow Possible"
    WINTRY_MIX_LIKELY = "Wintry Mix Likely"
    WINTRY_MIX_POSSIBLE = "Wintry Mix Possible"
    THUNDERSTORMS_LIKELY = "Thunderstorms Likely"
    THUNDERSTORMS_POSSIBLE = "Thunderstorms Possible"
    WINDY = "Windy"
    FOGGY = "Foggy"
    CLOUDY = "Cloudy"
    PARTLY_CLOUDY = "Partly Cloudy"
    VERY_LIGHT_RAIN = "Very Light Rain"
    LIGHT_RAIN = "Light Rain"
    MODERATE_RAIN = "Moderate Rain"
    HEAVY_RAIN = "Heavy Rain"
    VERY_HEAVY_RAIN = "Very Heavy Rain"
    EXTREME_RAIN = "Extreme Rain"


class Icon(Enum):
    CLEAR_DAY = "clear-day"
    CLEAR_NIGHT = "clear-night"
    CLOUDY = "cloudy"
    FOGGY = "foggy"
    PARTLY_CLOUDY_DAY = "partly-cloudy-day"
    PARTLY_CLOUDY_NIGHT = "partly-cloudy-night"
    POSSIBLY_RAINY_DAY = "possibly-rainy-day"
    POSSIBLY_RAINY_NIGHT = "possibly-rainy-night"
    POSSIBLY_SLEET_DAY = "possibly-sleet-day"
    POSSIBLY_SLEET_NIGHT = "possibly-sleet-night"
    POSSIBLY_SNOW_DAY = "possibly-snow-day"
    POSSIBLY_SNOW_NIGHT = "possibly-snow-night"
    POSSIBLY_THUNDERSTORM_DAY = "possibly-thunderstorm-day"
    POSSIBLY_THUNDERSTORM_NIGHT = "possibly-thunderstorm-night"
    RAINY = "rainy"
    SLEET = "sleet"
    SNOW = "snow"
    THUNDERSTORM = "thunderstorm"
    WINDY = "windy"

    @property
    def emoji(self) -> str:
        icon_mapping = {
            Icon.CLEAR_DAY: "☀️",
            Icon.CLEAR_NIGHT: "🌌️",
            Icon.CLOUDY: "☁️",
            Icon.FOGGY: "🌫️",
            Icon.PARTLY_CLOUDY_DAY: "⛅️",
            Icon.PARTLY_CLOUDY_NIGHT: "☁️",
            Icon.POSSIBLY_RAINY_DAY: "🌧️",
            Icon.POSSIBLY_RAINY_NIGHT: "🌧️",
            Icon.POSSIBLY_SLEET_DAY: "🌨️",
            Icon.POSSIBLY_SLEET_NIGHT: "🌨️",
            Icon.POSSIBLY_SNOW_DAY: "❄️",
            Icon.POSSIBLY_SNOW_NIGHT: "❄️",
            Icon.POSSIBLY_THUNDERSTORM_DAY: "⛈️",
            Icon.POSSIBLY_THUNDERSTORM_NIGHT: "⛈️",
            Icon.RAINY: "🌧️",
            Icon.SLEET: "🌨️",
            Icon.SNOW: "❄️️",
            Icon.THUNDERSTORM: "⛈️",
            Icon.WINDY: "💨️",
        }
        return icon_mapping.get(self, "❓️")

    @property
    def ha_icon(self) -> str:
        icon_mapping = {
            Icon.CLEAR_DAY: "sunny",
            Icon.CLEAR_NIGHT: "clear-night",
            Icon.CLOUDY: "cloudy",
            Icon.FOGGY: "fog",
            Icon.PARTLY_CLOUDY_DAY: "partlycloudy",
            Icon.PARTLY_CLOUDY_NIGHT: "partlycloudy",
            Icon.POSSIBLY_RAINY_DAY: "pouring",
            Icon.POSSIBLY_RAINY_NIGHT: "pouring",
            Icon.POSSIBLY_SLEET_DAY: "hail",
            Icon.POSSIBLY_SLEET_NIGHT: "hail",
            Icon.POSSIBLY_SNOW_DAY: "snowy",
            Icon.POSSIBLY_SNOW_NIGHT: "snowy",
            Icon.POSSIBLY_THUNDERSTORM_DAY: "lightning-rainy",
            Icon.POSSIBLY_THUNDERSTORM_NIGHT: "lightning-rainy",
            Icon.RAINY: "rainy",
            Icon.SLEET: "hail",
            Icon.SNOW: "snowy",
            Icon.THUNDERSTORM: "lightning",
            Icon.WINDY: "windy-variant",
        }
        return icon_mapping.get(self, "exceptional")


class PrecipType(Enum):
    RAIN = "rain"
    SNOW = "snow"
    SLEET = "sleet"
    STORM = "storm"
    NONE = "none"


class PrecipIcon(Enum):
    CHANCE_RAIN = "chance-rain"
    CHANCE_SNOW = "chance-snow"
    CHANCE_SLEET = "chance-sleet"
    NONE = "none"


class PressureTrend(Enum):
    FALLING = "falling"
    STEADY = "steady"
    RISING = "rising"
    UNKNOWN = "unknown"


class WindDirection(Enum):
    N = "N"
    NNE = "NNE"
    NE = "NE"
    ENE = "ENE"
    E = "E"
    ESE = "ESE"
    SE = "SE"
    SSE = "SSE"
    S = "S"
    SSW = "SSW"
    SW = "SW"
    WSW = "WSW"
    W = "W"
    WNW = "WNW"
    NW = "NW"
    NNW = "NNW"


class TemperatureUnit(Enum):
    CELSIUS = "C"
    FAHRENHEIT = "F"


class WindSpeedUnit(Enum):
    METERS_PER_SECOND = "m/s"
    KILOMETERS_PER_HOUR = "km/h"
    MILES_PER_HOUR = "mph"


class VisibilityUnit(Enum):
    KILOMETERS = "km"
    MILES = "mi"


@dataclass_json()
@dataclass(frozen=True, eq=True)
class CurrentConditions:
    air_temperature: float
    conditions: Condition

    feels_like: float
    icon: Icon  # Added missing field
    is_precip_local_day_rain_check: bool

    pressure_trend: PressureTrend
    relative_humidity: int
    sea_level_pressure: float
    solar_radiation: int
    time: int
    uv: int
    wet_bulb_globe_temperature: float
    wet_bulb_temperature: float
    wind_avg: float
    wind_direction: float
    wind_direction_cardinal: WindDirection
    wind_gust: float

    brightness: int | None = None
    station_pressure: float | None = None
    delta_t: float | None = None
    dew_point: float | None = None
    lightning_strike_count_last_1hr: int = 0
    lightning_strike_count_last_3hr: int = 0
    lightning_strike_last_distance: int = 0
    lightning_strike_last_distance_msg: str = ""
    lightning_strike_last_epoch: int | None = None
    precip_accum_local_day: int | None = None
    precip_accum_local_yesterday: int | None = None
    precip_minutes_local_day: int | None = None
    precip_minutes_local_yesterday: int | None = None
    air_density: float | None = None
    is_precip_local_yesterday_rain_check: bool | None = None


@dataclass_json
@dataclass(frozen=True, eq=True)
class ForecastDaily:
    air_temp_high: float
    air_temp_low: float
    conditions: Condition
    day_num: int
    day_start_local: int
    icon: Icon
    month_num: int
    precip_probability: int
    sunrise: int
    sunset: int

    precip_icon: PrecipIcon = PrecipIcon.NONE
    precip_type: PrecipType = PrecipType.NONE

    @property
    def rfc3939_datetime(self):
        utc_datetime = datetime.utcfromtimestamp(self.day_start_local)
        return utc_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")

    @property
    def ha_forecast(self) -> dict:
        return {
            "datetime": self.rfc3939_datetime,
            "is_daytime": None,
            "cloud_coverage": None,
            "condition": self.icon.ha_icon,
            "humidity": None,
            "native_apparent_temperature": None,
            "native_dew_point": None,
            "native_precipitation": None,
            "native_pressure": None,
            "native_temperature": self.air_temp_high,
            "native_templow": self.air_temp_low,
            "native_wind_gust_speed": None,
            "native_wind_speed": None,
            "precipitation_probability": self.precip_probability,
            "uv_index": None,
            "wind_bearing": None,
        }


@dataclass_json
@dataclass(frozen=True, eq=True)
class ForecastHourly:
    air_temperature: float
    conditions: Condition
    feels_like: float
    icon: Icon
    local_day: int
    local_hour: int
    precip: int
    precip_probability: int
    precip_type: PrecipType
    relative_humidity: int
    sea_level_pressure: float
    time: int
    uv: float
    wind_avg: float
    wind_direction_cardinal: WindDirection
    wind_direction: float
    wind_gust: float

    @property
    def rfc3939_datetime(self):
        utc_datetime = datetime.utcfromtimestamp(self.time)
        return utc_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")

    @property
    def ha_forecast(self) -> dict:
        """Property for Home Assistant to use"""
        return {
            # UTC Date time in RFC 3339 format.
            "datetime": self.rfc3939_datetime,
            "condition": self.icon.ha_icon,
            "humidity": self.relative_humidity,
            "native_apparent_temperature": self.feels_like,
            "native_precipitation": self.precip,
            "native_pressure": self.sea_level_pressure,
            "native_temperature": self.air_temperature,
            "native_wind_gust_speed": int(self.wind_gust),
            "native_wind_speed": self.wind_avg,
            "precipitation_probability": self.precip_probability,
            "uv_index": self.uv,
            "wind_bearing": self.wind_direction,
        }


@dataclass_json
@dataclass(frozen=True, eq=True)
class Forecast:
    daily: List[ForecastDaily]
    hourly: List[ForecastHourly]


@dataclass_json
@dataclass(frozen=True, eq=True)
class ForecastUnits:
    units_air_density: str
    units_brightness: str
    units_distance: str
    units_other: str
    units_precip: str
    units_pressure: str
    units_solar_radiation: str
    units_temp: str
    units_wind: str


@dataclass_json
@dataclass(frozen=True, eq=True)
class WeatherDataForecastREST:
    current_conditions: CurrentConditions
    forecast: Forecast
    latitude: float
    longitude: float
    location_name: str
    timezone: str
    timezone_offset_minutes: int
    units: ForecastUnits



# models/unified_station.py
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional
from dataclasses_json import dataclass_json

from weatherflow4py.models.rest.status import Status


@dataclass_json
@dataclass(frozen=True, order=True)
class DeviceMeta:
    agl: float
    environment: str
    name: str
    wifi_network_name: str


@dataclass_json
@dataclass(frozen=True, order=True)
class DeviceSettings:
    show_precip_final: bool


@dataclass_json
@dataclass(frozen=True, order=True)
class Device:
    device_id: int
    device_meta: DeviceMeta
    device_type: str
    firmware_revision: str
    hardware_revision: str
    serial_number: str
    device_settings: Optional[DeviceSettings] = None


@dataclass_json
@dataclass(frozen=True, order=True)
class StationItem:
    item: str
    location_id: int
    location_item_id: int
    station_id: int
    station_item_id: int
    sort: Optional[int] = None
    device_id: Optional[int] = field(default = -1)


@dataclass_json
@dataclass(frozen=True, order=True)
class StationMeta:
    elevation: float
    share_with_wf: bool
    share_with_wu: bool


@dataclass_json
@dataclass(frozen=True, order=True)
class Capability:
    capability: str
    device_id: int
    environment: str
    agl: Optional[float] = None
    show_precip_final: Optional[bool] = None


@dataclass_json
@dataclass(frozen=True, order=True)
class Stations:
    created_epoch: int
    devices: list[Device]
    is_local_mode: bool
    last_modified_epoch: int
    latitude: float
    longitude: float
    name: str
    public_name: str
    station_id: int
    station_items: list[StationItem]
    station_meta: StationMeta
    timezone: str
    timezone_offset_minutes: int
    capabilities: list[Capability] | None = None  # Optional field for capabilities

    @property
    def outdoor_devices(self) -> list[Device]:
        """Return a list of outdoor devices."""
        return [d for d in self.devices if d.device_type == "ST"]

    @property
    def indoor_devices(self) -> list[Device]:
        """Return a list of indoor devices."""
        return [d for d in self.devices if d.device_type != "ST"]

    @property
    def elevation(self) -> float:
        return self.station_meta.elevation


@dataclass_json
@dataclass(frozen=True, order=True)
class StationsResponseREST:
    stations: list[Stations]
    status: Status


    @property
    def station_map(self) -> dict[int, Stations]:
        """Return a dictionary of stations."""
        return {station.station_id: station for station in self.stations}

    @property
    def station_device_map(self) -> dict[int: list[int]]:
        """Return a dictionary of station_id to list of device_ids."""
        return {station.station_id: [device.device_id for device in station.devices] for station in self.stations}

    @property
    def station_outdoor_device_map(self) -> dict[int: list[int]]:
        """Return a dictionary of station_id to list of outdoor device_ids."""
        return {station.station_id: [device.device_id for device in station.outdoor_devices] for station in self.stations}


    @property
    def device_station_map(self) -> dict[int, int]:
        """Return a dictionary of device_id to station_id."""
        return {device.device_id: station.station_id for station in self.stations for device in station.devices}

